#!/usr/bin/perl
#
#####  $Id: synchro_ldap.pl,v 1.6 2003/06/23 16:00:07 bzizou Exp $  #####

# This script is aimed to synchronise the hadmin database with the ldap
# directory.
# If a user exists or misses into the ldap directory, it is created or deleted
# into the SQL database of the SLIS interface. This is the opposite of the 
# "valid" script.

# This script is part of the SLIS Project initiated by the CARMI-Internet
# (Acadmie de Grenoble - France 38).
# Ce script fait partie du projet SLIS dmarr par le CARMI-Internet
# (Acadmie de Grenoble - France 38).
#
# This script have been writen by Bruno Bzeznik - copyright (C) 2003
# Ce script  t crit par Bruno Bzeznik - copyright (C) - 2003
#
# SLIS : Serveur de communications Linux pour l'Internet Scolaire.
# Copyright (C) 1998-2003 Bruno Bzeznik

use DBI;
use Sys::Syslog qw(:DEFAULT setlogsock);
use MIME::Base64;
use Net::LDAP;
use Unicode::String qw(latin1 utf8);


# Misc variables
do '/home/hadmin/slis.conf.pl';

# Load the SLIS include library
do ("$SLIS_BINDIR/slis-sys.inc.pl");

# SQL Connection variables
do '/home/hadmin/cgi-bin/base/config_pgsql.inc.pl';
$DRIVER="Pg";
$PORT="5432";

# Open a channel to syslog
setlogsock('unix');
openlog('synchro_ldap','pid',$LOG_FACILITY);

# SQL and LDAP connection
$dsn = "DBI:$DRIVER:dbname=$BASE_HADMIN;host=$PGSQL_SERVER;port=$PORT";
$dbh = DBI->connect($dsn,$ADMINUSER,$ADMINPASS);
$ldap = Net::LDAP->new($LDAP_SERVER);
if (!$ldap) {
  print "ERR: Could not connect to LDAP server $LDAP_SERVER!\n";
  syslog('err',"ERR: Could not connect to LDAP server $LDAP_SERVER!");
  exit 1;
}
$mesg = $ldap->bind();  
if ( $mesg->code ) {
  print "ERR: Could not bind to LDAP server $LDAP_SERVER!\n";
  syslog('err',"ERR: Could not bind to LDAP server $LDAP_SERVER!");
  exit 1;
}

# MAIN

# PART 1: Delete the SQL users that are not in the LDAP directory
#################################################################

# Select the SQL users
$sth = $dbh->prepare("SELECT id,login FROM users where tovalid='0'");
$sth->execute;
log_err($sth,'select');

# For each user of the SQL database
while ((@line=$sth->fetchrow_array) | log_err($sth,'fetchrow')) {
   $id=$line[0];
   $login=$line[1];
   
   # Search him into the LDAP directory
   $mesg = $ldap->search(
                       base   => $LDAP_USERS,
                       filter => "uid=$login",
                      );
   if ($mesg->code) {log_msg(ldap_error_text($mesg->code));}

   # If not found, delete it from the database
   if ($mesg->count < 1) {
     syslog('info',"Deleting $login from SQL base.");
     $sth2 = $dbh->prepare("DELETE from users where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from users');
     $sth2 = $dbh->prepare("DELETE from oldvalue where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from oldvalue');
     $sth2 = $dbh->prepare("DELETE from password where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from password');
     $sth2 = $dbh->prepare("DELETE from alias where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from alias');
     $sth2 = $dbh->prepare("DELETE from groups where userid=$id");
     $sth2->execute;
     log_err($sth2,'delete from groups');
   }
}

# PART 2: Update or create the SQL users from the LDAP directory
#################################################################

# Select the LDAP users
$mesg = $ldap->search(
                       base   => $LDAP_USERS,
                       filter => "uid=*",
                      );
if ($mesg->code) {log_msg(ldap_error_text($mesg->code));}

# For each user...
while(my $entry = $mesg->shift_entry) {
  # Get the common attributes
  $val=$entry->get('uid'); $uid=@$val[0];
  $val=$entry->get('cn'); $cn=@$val[0];
  $cn = utf8($cn)->latin1;
  $val=$entry->get('sn'); $sn=@$val[0];
  $sn = utf8($sn)->latin1;
  $givenname=$cn;
  $givenname=~s/ $sn//;
  $val=$entry->get('gidNumber'); $gidNumber=@$val[0];
  $val=$entry->get('gecos'); $gecos=@$val[0];
  $val=$entry->get('employeeNumber'); $employeeNumber=@$val[0];
  # Keep in mind an associative array of primary groups, for later use in part 3
  $primary_group{$uid}=$gidNumber;

  # Search the user into the SQL database
  $sth = $dbh->prepare("SELECT id,nom,prenom,login,infos,code FROM users where tovalid='0' and login='$uid'");
  $sth->execute;
  log_err($sth,'select');
  @line=$sth->fetchrow_array;
  log_err($sth,'fetchrow');

  # If the user is found, update it
  if ($#line>0) {
    $id=$line[0];
    $nom=$line[1];
    $prenom=$line[2];
    $login=$line[3];
    $infos=$line[4];
    $code=$line[5];

    # Update only if necessary
    if ($nom ne $sn || $prenom ne $givenname || $infos ne $gecos
        || $code ne $employeeNumber) {
      syslog('info',"Updating $login: \"$nom\"->\"$sn\", \"$prenom\"->\"$givenname\", \"$infos\"->\"$gecos\", \"$code\"->\"$employeeNumber\"");

      $sth2 = $dbh->prepare("
        UPDATE users set
        nom='$sn',
        prenom='$givenname',
        infos='$gecos',
        code='$employeeNumber'
        WHERE id='$id';
      ");
      $sth2->execute;
      log_err($sth2,'update');
    }
  }

  # Else, create it
  else {
    syslog('info',"Creating $uid into SQL base");

    # Insertion into users table
    $sth2 = $dbh->prepare("
        INSERT INTO users
        (login,nom,prenom,infos,code,passwd,tovalid,modif,nouveau,delete)
        VALUES
        ('$uid','$sn','$givenname','$gecos','$employeeNumber','md5','f','f','f','f');
    ");
    $sth2->execute;
    log_err($sth2,'insert');

    # Get the id of the new user
    $sth2 = $dbh->prepare("SELECT id from users where login='$uid';");
    $sth2->execute;
    log_err($sth2,'select');
    @line=$sth2->fetchrow_array;
    log_err($sth2,'fetchrow');
    $id=$line[0];  

    # Insertion into the password table 
    $sth2 = $dbh->prepare("INSERT INTO password (id,md5,crypt) VALUES ('$id','ldap','ldap');");
    $sth2->execute;
    log_err($sth2,'insert');

  }
}

# PART 3: Delete the SQL groups that are not in the LDAP directory
##################################################################

# Select the SQL groups
$sth = $dbh->prepare("SELECT id,nom FROM classes where tovalid='0'");
$sth->execute;
log_err($sth,'select');

# For each group of the SQL database
while ((@line=$sth->fetchrow_array) | log_err($sth,'fetchrow')) {
   $id=$line[0];
   $group=$line[1];

   # Search it into the LDAP directory
   $mesg = $ldap->search(
                       base   => $LDAP_GROUPS,
                       filter => "cn=$group",
                      );
   if ($mesg->code) {log_msg(ldap_error_text($mesg->code));}

   # If not found, delete it from the database
   if ($mesg->count < 1) {
     syslog('info',"Deleting group $group from SQL base.");
     $sth2 = $dbh->prepare("DELETE from classes where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from classes');
     $sth2 = $dbh->prepare("DELETE from groups where classeid=$id");
     $sth2->execute;
     log_err($sth2,'delete from groups');
     $sth2 = $dbh->prepare("DELETE from alias where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from alias');
     $sth2 = $dbh->prepare("DELETE from oldvalue where id=$id");
     $sth2->execute;
     log_err($sth2,'delete from oldvalue');
   }
}

# PART 4: Create SQL groups that are missing and synchronize allocations
########################################################################

# Select the LDAP groups
$mesg = $ldap->search(
                       base   => $LDAP_GROUPS,
                       filter => "gidNumber=*",
                      );
if ($mesg->code) {log_msg(ldap_error_text($mesg->code));}

# Select the current SQL group allocations
$sth = $dbh->prepare("
    SELECT users.login,classes.nom,\"primary\",users.id,classes.id
    FROM users,classes,groups
    WHERE users.id=groups.userid and classes.id=groups.classeid
");
$sth->execute;
while ((@line=$sth->fetchrow_array) | log_err($sth,'fetchrow')) {
   $sqlgroup{$line[0].":".$line[1]}=$line[2];
   $sqluserid{$line[0].":".$line[1]}=$line[3];
   $sqlclasseid{$line[0].":".$line[1]}=$line[4];
}

# For each LDAP group...
while( my $entry = $mesg->shift_entry) {
  # Get the common attributes
  $val=$entry->get('cn'); $cn=@$val[0];
  $val=$entry->get('gidNumber'); $gidNumber=@$val[0];
  $members=$entry->get('memberUid');


  # Search the group into the SQL database
  $sth2 = $dbh->prepare("SELECT id FROM classes where tovalid='0' and nom='$cn'");
  $sth2->execute;
  log_err($sth2,'select');
  @line=$sth2->fetchrow_array;
  log_err($sth2,'fetchrow');


  # If the group is not found, create it
  if ($line[0]) {
    $classeid=$line[0];
  }
  else {
    syslog('info',"Creating missing SQL group $cn.");
    $sth2 = $dbh->prepare("
       INSERT INTO classes (nom,modif,nouveau,delete,tovalid)
       VALUES ('$cn','f','f','f','f');
    ");
    $sth2->execute;
    log_err($sth2,'insert into classes');
    $sth2 = $dbh->prepare("SELECT id FROM classes where tovalid='0' and nom='$cn'");
    $sth2->execute;
    log_err($sth2,'select');
    @line=$sth2->fetchrow_array;
    log_err($sth2,'fetchrow');
    $classeid=$line[0]; 
  }
 
  # Synchro  

  # Users that have this group as primary group
  while(($uid,$GID)=each %primary_group) {
      if ($GID == $gidNumber) {
        # If allocation does not exists, create it
        if ($sqlgroup{"$uid:$cn"} eq "") {
          syslog('info',"Allocating $uid to $cn as primary.");

          $sth3 = $dbh->prepare("SELECT id from users where login='$uid';");
          $sth3->execute;
          log_err($sth3,'select users');
          @line=$sth3->fetchrow_array;
          log_err($sth3,'fetchrow users.id');
          $userid=$line[0];
          $sth3 = $dbh->prepare("
             INSERT INTO groups 
             (classeid,userid,\"primary\",modif,nouveau,delete,tovalid) 
             VALUES
             ('$classeid','$userid','t','f','f','f','f');
          ");
          $sth3->execute;
          log_err($sth3,'insert groups');
        }
        # If allocation exists but not as primary, update it
        elsif ($sqlgroup{"$uid:$cn"} eq "0") {
          syslog('info',"Updating primary flag for $uid to $cn allocation.");
          $userid=$sqluserid{"$uid:$cn"};
          $sth3 = $dbh->prepare("UPDATE groups SET \"primary\"='t' where classeid='$classeid' and userid='$userid'");
          $sth3->execute;
          log_err($sth3,'update groups');
        }
      }
  }

  # Users that are members of this group (as secondary)
  foreach my $uid (@$members) {
    if ($primary_group{$uid} ne $gidNumber) {
        # If allocation does not exists, create it
        if ($sqlgroup{"$uid:$cn"} eq "") {
          $sth3 = $dbh->prepare("SELECT id from users where login='$uid';");
          $sth3->execute;
          log_err($sth3,'select users');
          @line=$sth3->fetchrow_array;
          log_err($sth3,'fetchrow users.id');
          if ($line[0]) {
            syslog('info',"Allocating $uid to $cn");
            $userid=$line[0];
            $sth3 = $dbh->prepare("
             INSERT INTO groups
             (classeid,userid,\"primary\",modif,nouveau,delete,tovalid)
             VALUES
             ('$classeid','$userid','t','f','f','f','f');
            ");
            $sth3->execute;
            log_err($sth3,'insert groups');
          }
        }
        # If allocation exists but as primary, update it
        elsif ($sqlgroup{"$uid:$cn"} eq "1") {
          syslog('info',"Updating primary flag for $uid to $cn allocation.");
          $userid=$sqluserid{"$uid:$cn"};
          $sth3 = $dbh->prepare("UPDATE groups SET \"primary\"='f' where classeid='$classeid' and userid='$userid'");
          $sth3->execute;
          log_err($sth3,'update groups');
        }
    }
  }   
}
